"""Tests for viz module - visualization utility functions."""

import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from matviz.viz import (
    plot_range, plot_cdf, plot_diag, plot_zero, 
    fancy_plotter, cplot, nicefy, subplotter
)


class TestPlotUtilities:
    """Test basic plotting utility functions."""

    def test_plot_cdf(self, sample_data):
        """Test plot_cdf function."""
        plot_cdf(sample_data['normal'])
        plt.close()
        
        # Should create a plot without error
        assert True

    def test_plot_cdf_with_nans(self):
        """Test plot_cdf handles NaN values."""
        data = np.array([1, 2, np.nan, 4, 5, np.nan, 7, 8])
        plot_cdf(data)
        plt.close()
        
        assert True

    def test_plot_diag(self):
        """Test plot_diag function."""
        plt.figure()
        plt.xlim(0, 10)
        plt.ylim(0, 10)
        plot_diag()
        plt.close()
        
        assert True

    def test_plot_zero(self):
        """Test plot_zero function."""
        plt.figure()
        plt.xlim(0, 10)
        plt.ylim(0, 10)
        plot_zero(lineheight=5, axx='x')
        plt.close()
        
        assert True


class TestComplexPlotting:
    """Test complex number plotting functions."""

    def test_cplot(self):
        """Test cplot function with complex data."""
        np.random.seed(42)
        z = np.random.randn(100) + 1j * np.random.randn(100)
        cplot(z, 'o')
        plt.close()
        
        assert True


class TestFancyPlotter:
    """Test fancy_plotter function."""

    def test_fancy_plotter_basic(self):
        """Test fancy_plotter with basic data."""
        x = np.linspace(0, 10, 20)
        y = 2 * x + 1 + np.random.randn(20) * 0.5
        
        fancy_plotter(x, y)
        plt.close()
        
        assert True

    def test_fancy_plotter_with_nans(self):
        """Test fancy_plotter handles NaN values."""
        x = np.array([1, 2, np.nan, 4, 5])
        y = np.array([2, 4, np.nan, 8, 10])
        
        fancy_plotter(x, y)
        plt.close()
        
        assert True


class TestSubplotter:
    """Test subplot functions."""

    def test_subplotter_basic(self):
        """Test basic subplotter functionality."""
        plt.figure(figsize=(10, 8))
        
        ax = subplotter(2, 2, 0)
        assert ax is not None
        
        ax = subplotter(2, 2, 1)
        assert ax is not None
        
        plt.close()


class TestStyling:
    """Test styling functions."""

    def test_nicefy_basic(self):
        """Test nicefy function basic functionality."""
        plt.figure()
        plt.plot([1, 2, 3], [1, 4, 9])
        plt.xlabel("X axis")
        plt.ylabel("Y axis")
        plt.title("Test plot")
        
        nicefy()
        plt.close()
        
        # Check that spines are modified
        assert True


#
# Smallest enclosing circle - Library (Python)
#
# Copyright (c) 2020 Project Nayuki
# https://www.nayuki.io/page/smallest-enclosing-circle
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program (see COPYING.txt and COPYING.LESSER.txt).
# If not, see <http://www.gnu.org/licenses/>.
#

import math, random
import numpy as np
from scipy import spatial

# Data conventions: A point is a pair of floats (x, y). A circle is a triple of floats (center x, center y, radius).

# Returns the smallest circle that encloses all the given points. Runs in expected O(n) time, randomized.
# Input: A sequence of pairs of floats or ints, e.g. [(0,5), (3.1,-2.7)].
# Output: A triple of floats representing a circle.
# Note: If 0 points are given, None is returned. If 1 point is given, a circle of radius 0 is returned.
#
# Initially: No boundary points known
def remove_nans(z):
    # remove nans
    I = np.isnan(z)
    return z[~I]

def cget_area(z):
    """
    Compute the convex hull area of complex-valued points.

    Parameters
    ----------
    z : complex array-like
        Points as complex numbers (real=x, imag=y). NaN values
        are removed.

    Returns
    -------
    float
        Convex hull area, or 0 if fewer than 4 unique points.
    """
    z = remove_nans(z)

    # you need at least three different points to get an area
    if len(np.unique(z)) > 3:
        points = [(np.real(zn), np.imag(zn)) for zn in z]
        hull = spatial.ConvexHull(points)
        area = hull.area
    else:
        area = 0
    return area


def cmake_circle2(z):
    """
    Find the smallest enclosing circle, returning [x, y, r].

    Parameters
    ----------
    z : complex array-like
        Points as complex numbers. NaN values are removed.

    Returns
    -------
    list of [float, float, float]
        ``[center_x, center_y, radius]``.
    """
    # remove nans
    z = remove_nans(z)

    if len(np.unique(z)) > 3:
        # convert to (x,y)
        points = [(np.real(zn), np.imag(zn)) for zn in z]
        [x, y, r] = make_circle(points)
    else:
        x = np.nan
        y = np.nan
        r = np.nan

    return [x, y, r]

def cmake_circle(z):
    """
    Find the smallest enclosing circle, returning complex center and radius.

    Parameters
    ----------
    z : complex array-like
        Points as complex numbers. NaN values are removed.

    Returns
    -------
    z_center : complex
        Center of the enclosing circle.
    r : float
        Radius.
    """
    # convert x, y back into z for the center
    [x, y, r] = cmake_circle2(z)
    z_center = x + 1j * y
    return z_center, r


def make_circle(points):
    # Convert to float and randomize order
    shuffled = [(float(x), float(y)) for (x, y) in points]
    random.shuffle(shuffled)

    # Progressively add points to circle or recompute circle
    c = None
    for (i, p) in enumerate(shuffled):
        if c is None or not is_in_circle(c, p):
            c = _make_circle_one_point(shuffled[: i + 1], p)
    return c


# One boundary point known
def _make_circle_one_point(points, p):
    c = (p[0], p[1], 0.0)
    for (i, q) in enumerate(points):
        if not is_in_circle(c, q):
            if c[2] == 0.0:
                c = make_diameter(p, q)
            else:
                c = _make_circle_two_points(points[: i + 1], p, q)
    return c


# Two boundary points known
def _make_circle_two_points(points, p, q):
    circ = make_diameter(p, q)
    left = None
    right = None
    px, py = p
    qx, qy = q

    # For each point not in the two-point circle
    for r in points:
        if is_in_circle(circ, r):
            continue

        # Form a circumcircle and classify it on left or right side
        cross = _cross_product(px, py, qx, qy, r[0], r[1])
        c = make_circumcircle(p, q, r)
        if c is None:
            continue
        elif cross > 0.0 and (
                left is None or _cross_product(px, py, qx, qy, c[0], c[1]) > _cross_product(px, py, qx, qy, left[0],
                                                                                            left[1])):
            left = c
        elif cross < 0.0 and (
                right is None or _cross_product(px, py, qx, qy, c[0], c[1]) < _cross_product(px, py, qx, qy, right[0],
                                                                                             right[1])):
            right = c

    # Select which circle to return
    if left is None and right is None:
        return circ
    elif left is None:
        return right
    elif right is None:
        return left
    else:
        return left if (left[2] <= right[2]) else right


def make_diameter(a, b):
    cx = (a[0] + b[0]) / 2
    cy = (a[1] + b[1]) / 2
    r0 = math.hypot(cx - a[0], cy - a[1])
    r1 = math.hypot(cx - b[0], cy - b[1])
    return (cx, cy, max(r0, r1))


def make_circumcircle(a, b, c):
    # Mathematical algorithm from Wikipedia: Circumscribed circle
    ox = (min(a[0], b[0], c[0]) + max(a[0], b[0], c[0])) / 2
    oy = (min(a[1], b[1], c[1]) + max(a[1], b[1], c[1])) / 2
    ax = a[0] - ox;
    ay = a[1] - oy
    bx = b[0] - ox;
    by = b[1] - oy
    cx = c[0] - ox;
    cy = c[1] - oy
    d = (ax * (by - cy) + bx * (cy - ay) + cx * (ay - by)) * 2.0
    if d == 0.0:
        return None
    x = ox + ((ax * ax + ay * ay) * (by - cy) + (bx * bx + by * by) * (cy - ay) + (cx * cx + cy * cy) * (ay - by)) / d
    y = oy + ((ax * ax + ay * ay) * (cx - bx) + (bx * bx + by * by) * (ax - cx) + (cx * cx + cy * cy) * (bx - ax)) / d
    ra = math.hypot(x - a[0], y - a[1])
    rb = math.hypot(x - b[0], y - b[1])
    rc = math.hypot(x - c[0], y - c[1])
    return (x, y, max(ra, rb, rc))


_MULTIPLICATIVE_EPSILON = 1 + 1e-14


def is_in_circle(c, p):
    return c is not None and math.hypot(p[0] - c[0], p[1] - c[1]) <= c[2] * _MULTIPLICATIVE_EPSILON


# Returns twice the signed area of the triangle defined by (x0, y0), (x1, y1), (x2, y2).
def _cross_product(x0, y0, x1, y1, x2, y2):
    return (x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)


import fitz  # PyMuPDF
import io
import os
from PIL import Image
import shutil


def extract_images_from_pdf(input_pdf_path: str, output_folder: str) -> str:
    """
    Extracts all images from the given PDF and saves them in the specified output folder.
    Returns the path to a zipped folder containing all extracted images.
    """
    doc = fitz.open(input_pdf_path)
    os.makedirs(output_folder, exist_ok=True)
    image_paths = []

    for page_index in range(len(doc)):
        page = doc[page_index]
        images = page.get_images(full=True)

        for img_index, img in enumerate(images):
            xref = img[0]
            base_image = doc.extract_image(xref)
            image_bytes = base_image["image"]
            image_ext = base_image["ext"]

            image_path = os.path.join(output_folder, f"page{page_index + 1}_img{img_index + 1}.{image_ext}")
            with open(image_path, "wb") as img_file:
                img_file.write(image_bytes)

            image_paths.append(image_path)

    doc.close()

    # Zip the folder for easier distribution
    zip_path = shutil.make_archive(output_folder, 'zip', output_folder)
    return zip_path


def compress_pdf_images(input_pdf_path: str, output_pdf_path: str, dpi: int = 150, quality: int = 40) -> str:
    """
    Compresses all images in the PDF by rendering pages as compressed images.
    Returns the path to the compressed PDF.

    dpi: controls the resolution of the page rendering.
    quality: JPEG quality setting (0-100).
    """
    doc = fitz.open(input_pdf_path)
    compressed_pdf = fitz.open()

    for page_index in range(len(doc)):
        page = doc[page_index]
        pix = page.get_pixmap(dpi=dpi)
        img = Image.frombytes("RGB", [pix.width, pix.height], pix.samples)
        img_byte_arr = io.BytesIO()
        img.save(img_byte_arr, format='JPEG', quality=quality)
        img_stream = img_byte_arr.getvalue()

        new_page = compressed_pdf.new_page(width=page.rect.width, height=page.rect.height)
        rect = fitz.Rect(0, 0, page.rect.width, page.rect.height)
        new_page.insert_image(rect, stream=img_stream)

    compressed_pdf.save(output_pdf_path)
    compressed_pdf.close()
    doc.close()

    return output_pdf_path



# Configuration file for the Sphinx documentation builder.

import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information
project = "matviz"
copyright = "2019, Jonathan C. Lansey"
author = "Jonathan C. Lansey"

# -- General configuration
extensions = [
    "myst_nb",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "numpydoc",
    "sphinx_copybutton",
    "sphinx_design",
]

templates_path = ["_templates"]
exclude_patterns = []

# MyST settings
myst_enable_extensions = [
    "colon_fence",
    "deflist",
]

# Notebook execution
nb_execution_mode = "off"

# Autodoc settings
autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "show-inheritance": True,
}

# Numpydoc settings
numpydoc_show_class_members = False

# Intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
}

# -- Options for HTML output
html_theme = "pydata_sphinx_theme"

html_theme_options = {
    "github_url": "https://github.com/JLansey/matviz",
    "navbar_align": "left",
    "navigation_with_keys": False,
}

html_static_path = ["_static"]

# Suppress warnings about missing references for built-in types
nitpicky = False
